"""A module for handling the information the player receives each round."""

from dataclasses import dataclass, field

@dataclass
class Hint:
    """A summary of the information the player receives each round."""

    correct: dict[str, set[int]] = field(default_factory=dict)
    inverse_correct: dict[int, str] = field(default_factory=dict, init=False)

    misplaced: dict[str, set[int]] = field(default_factory=dict)
    inverse_misplaced: dict[int, set[str]] = field(default_factory=dict, init=False)

    absent: set[str] = field(default_factory=set)

    def __post_init__(self):
        """Initialize inverse mappings."""
        for letter, indices in self.correct.items():
            for index in indices:
                self.inverse_correct[index] = letter

        for letter, indices in self.misplaced.items():
            for index in indices:
                self.inverse_misplaced.setdefault(index, set()).add(letter)



"""User interface."""

import colorama
from tap import Tap

from .game import Game

class _ArgumentParser(Tap):
    bot: bool = False  # have the bot play instead

    def configure(self):
        self.add_argument('-b', '--bot')

def cli():
    """Command-line interface."""
    args = _ArgumentParser(description="Interact with Wordle.").parse_args()

    colorama.init()
    Game(bot=args.bot).play()



"""A module for administering the information the player receives each round."""

from ..common.hint import Hint

def hint(answer: str, guess: str) -> Hint:
    """Hint at the answer based on a guess."""
    hint = Hint()
    working_answer = answer[:]

    # First, scan for correct letters
    for guess_letter_index, guess_letter in enumerate(guess):
        if working_answer[guess_letter_index] == guess_letter:
            # This letter is correct; register its location
            hint.correct.setdefault(guess_letter, set()).add(guess_letter_index)
            hint.inverse_correct[guess_letter_index] = guess_letter

            # Let's remove this letter from our working answer, so we don't
            # match its specific occurrence again when scanning for misplaced
            # and absent letters.
            #
            # We should replace it with an unmatchable character rather than
            # remove it outright, since we want to preserve the locations of
            # the remaining letters.
            working_answer = working_answer[:guess_letter_index] \
                           + ' ' \
                           + working_answer[guess_letter_index + 1:]

    # Now, scan for misplaced and absent letters
    for guess_letter_index, guess_letter in enumerate(guess):
        if guess_letter_index in hint.inverse_correct:
            # Skip correct letters
            continue

        try:
            answer_letter_index = working_answer.index(guess_letter)
        except ValueError:
            # This letter is absent; register it
            hint.absent.add(guess_letter)
            continue

        # This letter is misplaced; register its location
        hint.misplaced.setdefault(guess_letter, set()).add(guess_letter_index)
        hint.inverse_misplaced.setdefault(guess_letter_index, set()).add(guess_letter)

        # Again, prevent further matches
        working_answer = working_answer[:answer_letter_index] \
                       + ' ' \
                       + working_answer[answer_letter_index + 1:]

    return hint
